/************************************************************************ 
 *
 * 24Cxx version 1.0 - Copyright (C) 2000 Luca Ferretti <luca.ferretti@wind.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *            See the file COPYING for more details.
 *
 *************************************************************************
 *            	  
 * PROGRAM : 24Cxxx.c  
 * RELEASE : 1.0 
 * DATE    : 17 Sep 2000
 * PURPOSE : Read / Write / Verify the content of an i2c eeprom
 *           type 24Cxxx connected to a serial port via Ludipipo|JDM i/f
 * 
 ************************************************************************/	

#include "24Cxxx.h"

// Beg of main
int main(int argc,char *argv[]) {
	unsigned char buff[MAX_E2PROM_SIZE];
	unsigned char buff1[MAX_E2PROM_SIZE];
	char port[255]=DEFAULT_PORT;
	char file_name[255]=DEFAULT_FNAME;
	int i,eeprom_size=B_SIZE,verify_errors=0;
	char optchar;
	// Parse cmd line opts	
	if(argc<2)
		usage();
	opterr=0;
	while ((optchar = getopt(argc, argv, "f: r w v d: help s:")) != -1) {
		switch(optchar) {
			case 'w' :
				options|=W_BIT_MASK;
				continue;
			case 'r' :
				options|=R_BIT_MASK;
				if (optarg != 0) 
					eeprom_size=(atoi(optarg)*1024)/8;
				continue;
			case 'v' :
				options|=V_BIT_MASK;
				continue;
			case 'd' :
				strcpy(port,optarg);
				continue;
			case 'f' :
				strcpy(file_name,optarg);
				continue;
			case 's' :
				wct=(atoi(optarg));
				continue;
			case 'h' :
				usage();
			default :
				usage();
		}
	}
	if (optind<argc){ 			// set eeprom size if passed 
		eeprom_size=(atoi(argv[optind])*1024)/8;
		pg_size=get_pg_size(argv[optind]);
	}
	// end of parsing cmd line

	if (!(options & W_BIT_MASK || options & R_BIT_MASK || options & V_BIT_MASK))
		usage();
	
	if ((options & W_BIT_MASK) && (options & R_BIT_MASK)) {
		fprintf(stdout,"\nRead & Write @ the same time ?\n");
		usage();
	}

	fprintf(stdout,"%s",header);
	fd=open_port(port);

	schedulize_srt();			// let's go in soft RT mode	
	
	vcc_on;					// Power up IC ...
	usleep(POWERUP_DELAY);			// .. and wait a moment 
	
	if (options & R_BIT_MASK) {	
		read_e2prom(buff,eeprom_size);
		save_buffer(file_name,buff,eeprom_size);
	}
	if (options & W_BIT_MASK) {	
		eeprom_size=read_buffer(file_name,buff); 
		write_e2prom(buff,eeprom_size);
	}
	if (options & V_BIT_MASK) {
		fprintf(stdout,"Verify in progress...\n");
		if ( !(options & W_BIT_MASK) && !(options & R_BIT_MASK) )
			eeprom_size=read_buffer(file_name,buff);
		read_e2prom(buff1,eeprom_size);
		for (i=0;i<eeprom_size;i++) {
			if (buff[i] != buff1[i]) {
				verify_errors++;
				fprintf(stderr," ERROR @ location %d[%.4xh] \"%s\" data='%.2x' e2prom data='%.2x'\n",i,i,file_name,buff[i],buff1[i]); 
			}
		}
		if (verify_errors)
			fprintf(stdout,"Verify : %d errors found\n",verify_errors);
		else
			fprintf(stdout,"Verify : OK\n");
	}
	idle();
	close(fd);         // we don't need serial anymore
	fprintf(stdout,"\n---------------------------------------------------------------------------\n\n");
	return(0);
}
// End of main

// Funcs defs

void schedulize_srt(void) {
	struct sched_param sp;
	sp.sched_priority=sched_get_priority_max(SCHED_FIFO);
	sched_setscheduler(0,SCHED_FIFO,&sp);
}	

int open_port(char * port) { // Returns the file descriptor or die on error 
	int fd;  
	fd = open(port, O_RDWR ); 
	if (fd ==  -1) {  
		fprintf(stderr,"\nERROR: Unable to open device %s\n",port);
		exit(1);
	} else fcntl(fd, F_SETFL, 0); return (fd); 
}

void delay(int delay) {		
	while(delay--);
}

unsigned char recv_bit(void) {
	int status;
	scl_lo; delay(TXX);
	sda_hi; delay(TXX);
	scl_hi; delay(TXX);
	ioctl(fd, TIOCMGET, &status); 
	status= status & TIOCM_CTS;
	if ( status )
		return 0x01;
	else
		return 0x00;
}

void send_bit(unsigned char bit) {
	scl_lo; delay(TXX);
	if (bit) 
		sda_hi;
	else	
		sda_lo; 
	delay(TXX);
	scl_hi; delay(TXX);
}

void send_ack(void) {
	send_bit(0);
}

void idle(void) {
	sda_hi;
	scl_hi;
	vcc_off;
}

unsigned char recv_ack(void) {
	unsigned char ack;
	ack=recv_bit();
	scl_lo;
	if (ack) { 
		idle();
		close(fd);  
		fprintf(stderr,"\n\nACK ERROR !\n\n");
		fprintf(stderr,"Suggestion : increase -s value.\n\n");
		exit(3);
	}	
	return (ack);
}
		
void send_start(void) {
	sda_hi; delay(TXX);
	scl_hi; delay(TXX);
	sda_lo; delay(TXX);
}

void send_stop(void) {
	sda_lo; delay(TXX);
	scl_hi; delay(TXX);
	sda_hi; delay(TXX);
}
		
void send_byte(unsigned char byte) {
	int i;
	for (i=7;i>=0;i--) 				// Send
		send_bit((byte>>i)&0x01);		// MSb first
}

unsigned char recv_byte(void) {
	int i;
	unsigned char byte=0;
	for (i=7;i>=0;i--)				// receive
		byte|=recv_bit()<<i;			// MSb first
	return byte;
}
		
void read_e2prom(unsigned char *buffer, int size) {
	int i;
	fprintf(stdout,"Reading %d bytes...",size);	
	fflush(stdout);
	send_start(); send_byte(W_HEADER); 
	recv_ack();
	send_byte(0x00);	// let's start reading from addr 0
	recv_ack(); 
	send_start(); send_byte(R_HEADER);
	recv_ack();
	for (i=0;i<size;i++) {
		*buffer++=recv_byte();
		send_ack();
		progress();
	}
	send_stop();
	fprintf(stdout,"done\n");	
}

void save_buffer(char *f_name,unsigned char *buffer,int size) {
	int i;
	FILE *fp;

	fprintf(stdout,"Writing %d bytes to \"%s\"...",size,f_name);
	fflush(stdout);
	if (NULL == ( fp=fopen(f_name,"w+b") ) ) {
		fprintf(stderr,"\nERROR : cannot create %s\n\n",f_name);
		exit(1);
	}
	for(i=0;i<size;i++) 
		putc(*buffer++,fp);	
	fclose(fp);
	fprintf(stdout,"done\n");
}

int read_buffer(char  *f_name,unsigned char *buffer) {
	int i,fsize;
	FILE *fp;

	fprintf(stdout,"Reading \"%s\"...",f_name);
	fflush(stdout);
	if (NULL == (fp = fopen(f_name, "rb"))) {
		fprintf(stderr,"\n\nERROR : cannot find %s\n\n",f_name);
		exit(1);
	}
	fseek(fp, 0L, SEEK_END);		// Go 2 *real* EOF
	fsize=ftell(fp);			// now we have bytes GT
	fseek(fp,0L,SEEK_SET);			// Rewind 2 beg
	i=fsize;	
	while (i--) 
		*buffer++=getc(fp);
	fclose(fp);
	fprintf(stdout,"done (%d bytes)\n",fsize);
	return(fsize);
}
     
void write_e2prom(unsigned char *buffer,int size) {
	int i,k;
	fprintf(stdout,"Writing %d bytes...",size);
	fflush(stdout);
	for (i=0;i<size;i++) {
		send_start();
		if (size<=2048) 		// byte addr protocol
	       		send_byte(W_HEADER+((i/0x100)<<1)); // + 3b addr MSB
	       	else {				// word addr protocol
			send_byte(W_HEADER);
			recv_ack();
			send_byte(i/0x100);
		}	
		recv_ack();
		send_byte((i%0x100));		// Addr LSB
		recv_ack();
		for (k=0;k<pg_size;k++,i++) {  // page write
			send_byte(*buffer++);
			recv_ack();
		}
		i--;
		send_stop();
		progress();
		//delay(wct);
		usleep(wct);
	}
	fprintf(stdout,"done\n");
}

void progress(void) {
	static int progress_val;
	char whirl[4]={'-','\\','|','/'};
	fprintf(stdout,"%c\b",whirl[progress_val++]);
	fflush(stdout);
	if (progress_val==4)
		progress_val=0;
}

int get_pg_size(char *KB) {
	#define HASH_N 10 
	//char *ee[HASH_N] = { "1","2","4","8","16","32","64","128","256","512" };
	char *ee[HASH_N] = E2PROMS ;
	//int pp[HASH_N] = { 8,8,16,16,16,32,32,32,32,32 };
	int pp[HASH_N] = PG_SIZES;
	ENTRY ps, *psp;
	int i;
	hcreate(HASH_N);
	for (i = 0; i < HASH_N; i++)  {
		ps.key = ee[i];     
		ps.data = (char *) pp[i];
		psp=hsearch(ps,ENTER);  
	}
	ps.key = KB;
       	psp=hsearch(ps,FIND);
	if (psp)
	       	return((int)psp->data); 
	else 
		return(8);
}

void usage(void) {
	fprintf(stdout,"%s",header);	
	fprintf(stdout,"   Read/Program/Verify i2c eeproms via Ludipipo|JDM serial interface\n");
	fprintf(stdout,"\nUSAGE : 24Cxxx [SIZE] MODE [OPTIONS] \n\n");
	fprintf(stdout,"SIZE : 		set eeprom size to SIZE Kbits (default:16),\n");
	fprintf(stdout,"				needed for -r mode only\n");
	fprintf(stdout,"MODE :		-r		read from IC\n");
	fprintf(stdout,"		-w		write to IC\n");
	fprintf(stdout,"		-v		verify IC or file content\n");
	fprintf(stdout,"OPTIONS : 	-f filename	use filename (default:eeprom.bin)\n");
	fprintf(stdout,"		-d dev		use device dev for programmer\n");
	fprintf(stdout,"                                (default:/dev/tty0)\n");
	fprintf(stdout,"		-s val		set write cycle time to approx val us \n");
	fprintf(stdout,"				(default:10000 [10ms])\n\n");
	exit(99);
}
    
